/*
 * $Id: MultiPartyMixer.cpp,v 1.6 2003/11/25 16:19:18 rco Exp $
 *
 * Copyright (C) 2002-2003 Fhg Fokus
 *
 * This file is part of sems, a free SIP media server.
 *
 * sems is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * For a license to use the sems software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * sems is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "MultiPartyMixer.h"
#include "AmRtpStream.h"
#include "log.h"

#include <assert.h>

// PCM16 range: [-32767:32768]
#define MAX_LINEAR_SAMPLE 32737

MultiPartyMixer::MultiPartyMixer(AmRtpStreamSet* str_set)
    : channels(),ch_offsets(),
      cur_channel_id(0),
      scaling_factor(16),
      str_set(str_set)
{
}

MultiPartyMixer::~MultiPartyMixer()
{
}

unsigned int MultiPartyMixer::addChannel()
{
    for(;channels.find(cur_channel_id) != channels.end();cur_channel_id++)
	DBG("trying to add channel: #%i\n",cur_channel_id);

    DBG("added channel: #%i\n",cur_channel_id);
    channels.insert(std::make_pair(cur_channel_id,SampleArrayShort()));

    return cur_channel_id++;
}

void MultiPartyMixer::removeChannel(int channel_id)
{
    channels.erase(channel_id);
    ch_offsets.erase(channel_id);
    DBG("removed channel: #%i\n",channel_id);
}

SampleArrayShort* MultiPartyMixer::get_channel(unsigned int channel_id)
{
    ChannelMap::iterator channel_it = channels.find(channel_id);
    if(channel_it == channels.end()){
	ERROR("channel %i does not exist\n",channel_id);
	return NULL;
    }
    
    return &channel_it->second;
}

void MultiPartyMixer::PutChannelPacket(unsigned int   channel_id,
				       unsigned int   user_ts,
				       unsigned char* buffer, 
				       unsigned int   size)
{
    if(!size) return;
    assert(size <= AUDIO_BUFFER_SIZE);
    
    unsigned int ts_set = str_set->getTS();
    if(ch_offsets.find(channel_id) == ch_offsets.end())
      ch_offsets[channel_id] = ts_set - user_ts;

    unsigned int ts = ch_offsets[channel_id] + user_ts;

    if(cmp_ts(ts,ts_set-160*4)){

      unsigned int ts_set = str_set->getTS();
      ch_offsets[channel_id] = ts_set - user_ts;
      DBG("resynchronising channel %u (ts_set-ts=%u)\n",
	  channel_id,ts_set-ts);
      ts = ts_set;
    }
      
    if(cmp_ts(ts_set,ts)){

      unsigned int ts_set = str_set->getTS();
      ch_offsets[channel_id] = ts_set - user_ts;
      DBG("resynchronising channel %u (ts-ts_set=%u)\n",
	  channel_id,ts-ts_set);
      ts = ts_set;
    }

    if(SampleArrayShort* channel = get_channel(channel_id)){

	size >>= 1;
 	int* tmp_buf = new int[size];
 	try {

  	    channel->put(ts,(short*)buffer,size);

       	    mixed_channel.get(ts,tmp_buf,size);
       	    mix_add(tmp_buf,tmp_buf,(short*)buffer,size);
       	    mixed_channel.put(ts,tmp_buf,size);
 	}
 	catch(...){}

 	delete [] tmp_buf;
    }
}

void MultiPartyMixer::GetChannelPacket(unsigned int   channel_id,
				       unsigned int   user_ts,
				       unsigned char* buffer, 
				       unsigned int   size)
{
    assert(size <= AUDIO_BUFFER_SIZE);

    if(ch_offsets.find(channel_id) == ch_offsets.end())
      ch_offsets[channel_id] = str_set->getTS() - user_ts;

    unsigned int ts = ch_offsets[channel_id] + user_ts;

    SampleArrayShort* channel = 0;
    if((channel = get_channel(channel_id)) != 0){

	size >>= 1;
 	int* tmp_buf = new int[size];
 	try {
	    ts -= 160*4;
	    mixed_channel.get(ts,tmp_buf,size);
 	    channel->get(ts,(short*)buffer,size);
	    mix_sub(tmp_buf,tmp_buf,(short*)buffer,size);
     	    scale((short*)buffer,tmp_buf,size);
 	}
 	catch(...){}
	delete [] tmp_buf;
    }
}

// int   dest[size/2]
// int   src1[size/2]
// short src2[size/2]
//
void MultiPartyMixer::mix_add(int* dest,int* src1,short* src2,unsigned int size)
{
     int* end_dest = dest + size;

     while(dest != end_dest)
 	*(dest++) = *(src1++) + int(*(src2++));
}

void MultiPartyMixer::mix_sub(int* dest,int* src1,short* src2,unsigned int size)
{
     int* end_dest = dest + size;

     while(dest != end_dest)
 	*(dest++) = *(src1++) - int(*(src2++));
}

void MultiPartyMixer::scale(short* buffer,int* tmp_buf,unsigned int size)
{
    short* end_dest = buffer + size;
    
    if(scaling_factor<64)
	scaling_factor++;
    
    while(buffer != end_dest){
	
	int s = (*tmp_buf * scaling_factor) >> 6;
	if(abs(s) > MAX_LINEAR_SAMPLE){
	    scaling_factor = abs( (MAX_LINEAR_SAMPLE<<6) / (*tmp_buf) );
	    if(s < 0)
		s = -MAX_LINEAR_SAMPLE;
	    else
		s = MAX_LINEAR_SAMPLE;
	}
	*(buffer++) = short(s);
 	tmp_buf++;
    }
}
