/*
 * $Id: IsdnDeviceIncoming.cpp,v 1.3 2003/11/16 19:28:05 ullstar Exp $
 *
 * Copyright (C) 2003 Ulrich Abend
 *
 * This file is part of isdngw, a free isdn gateway plugin 
 * for sems, a free SIP media server.
 *
 * isdngw is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * isdngw is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "IsdnDevice.h"
#include "IsdnDeviceIncoming.h"
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <pthread.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "globals.h"
#include "IsdnConfiguration.h"
#include "IsdnChat.h"
#include "IsdnUtils.h"
#include "log.h"
#include "AmSession.h"

/** mutex is static in IsdnDeviceIncoming. */
pthread_mutex_t mut_reass;

/** Create a new incoming IsdnDevice. @see IsdnDevice(). */
IsdnDeviceIncoming::IsdnDeviceIncoming(char* devicename) 
 : IsdnDevice(devicename) 
{
	/* initialize mutex */
	pthread_mutex_init(&mut_state, NULL);
}

/** Called from Destructor. Closes the device, removes the lockfile and frees
 * all allocated ressources.
 */
void IsdnDeviceIncoming::cleanup() {
	pthread_mutex_unlock(&mut_state);
	pthread_mutex_destroy(&mut_state);
}

void IsdnDeviceIncoming::setCommand(AmCmd& cmd) {
	m_cmd = cmd;
}

void IsdnDeviceIncoming::onBeforeCallAccept(AmRequest* req)
{
	DBG("onBeforeCallAccept called for incoming call\n");
}	

void IsdnDeviceIncoming::onBye()
{
	ERROR("onBye called for incoming device\n");
	resetFailedIncomingDevice();
}

void IsdnDeviceIncoming::onError(unsigned int code, const string& reason)
{
	ERROR("could not setup call on SIP side\n");
	resetFailedIncomingDevice();
}

/** Acquire a locked device for inbound connection.
 * This function tries to find an unused incoming device.
 * @return a locked device on success <BR>
 * @return NULL if no device is available
 */
IsdnDevice* IsdnDeviceIncoming::getLockedDevice() {
	struct list_t* cur = isdn_config->i4l_indevices;
	int ok = FAIL;
	while (cur != NULL) {
		if ( (cur->num == STATUS_OK) && (((IsdnDevice*)(cur->link))->trylock() == OK) ) {
			ok = OK;
			break;
		} else {
			cur = cur->next;
		}
	}
	if (ok == OK) {
		return (IsdnDevice*)cur->link;
	} else {
		return NULL;
	}	
}

/** Answer an incoming call. 
 */
int IsdnDeviceIncoming::connectVoice() {
	DBG("connecting voice\n");
	pthread_mutex_lock(&mut_state);
	if (m_listening_state == LISTENING_STATE_RINGING) {
		DBG("call is still ringing, now setting to connected to end polling for RUNG\n");
		m_listening_state = LISTENING_STATE_CONNECTED;
	} else {	
		ERROR("call is not pending, device is resetted already, unlocking and leaving with FAIL, shouldn't happen\n");
		exit(-1);
	}
	pthread_mutex_unlock(&mut_state);

	IsdnChat ic = IsdnChat();
	ic.setAnswerIncomingCall();
	int res = ic.chat(m_fd);
	if (res == FAIL) {
		ERROR("call setup on pstn failed\n");
		resetFailedIncomingDevice();
		return FAIL;
	}	
	ic.reset();
	ic.setConnect();
	int ret = ic.chat(m_fd);
	if (ret == FAIL) {
		ERROR("call setup on pstn failed\n");
		resetFailedIncomingDevice();
		return FAIL;
	}	
	return ret;
}

/** Initialize the incoming device logic. Must be called exactly once 
 * before reassignIncomingDevices() is called the first time.
 */
/*static*/ void IsdnDeviceIncoming::initializeOnce() {
	pthread_mutex_init(&mut_reass, NULL);
}	

/** (Re)assign Devicethreads to incoming numbers. Every time a thread
 * stopps listening on a given number (e.g. if it is handling a call)
 * this method must be called. For each number, that does not have a
 * corresponding listener thread, one is spawned.
 */
void IsdnDeviceIncoming::reassignIncomingDevices() {
	pthread_mutex_lock(&mut_reass);
	struct list_t* cur = isdn_config->devices[DEVICE_I4L].inboundPatterns;
	while (cur != NULL) {
	//	DBG("searching assigned device for %s\n", cur->number);
		if (cur->link == NULL) { // no device assigned yet
		//	DBG("no device assigned yet\n");
			cur->link = (void*)getLockedDevice();
			if (cur->link != NULL) { // device available 
			//	DBG("acquired locked device :-)\n");
				((IsdnDeviceIncoming *)cur->link)->listenOnNumber(cur->item);
			} else {
				ERROR("could not get locked incoming device, number %s unhandled!\n", cur->item);
			}
		} else { // device already assigned
			// DBG("device already assigned\n");
			if (((IsdnDeviceIncoming*)cur->link)->m_listening_state != LISTENING_STATE_LISTENING) {
			//	DBG(" but not in listening state!\n");
				cur->link = (void*)getLockedDevice();
				if (cur->link != NULL) { // device available 
				//	DBG("assigning new device\n");
					((IsdnDeviceIncoming*)cur->link)->listenOnNumber(cur->item);
				} else {
					ERROR("could not get locked incoming device, number %s unhandled (2)!\n", cur->item);
				}
					
			}
		//	DBG("already listening on number %s\n", cur->number);
		}	
		cur = cur->next;
	}
	pthread_mutex_unlock(&mut_reass);
}	

/** Listen on a specified number, initializes the device and spawns 
 * a listening thread. Must be called with mut_reass already locked.
 */
pthread_t* IsdnDeviceIncoming::listenOnNumber(char* number) {
	m_incoming_number = number;
	IsdnChat ic = IsdnChat();
	ic.setIncomingDevice(number);
	if (ic.chat(m_fd) == FAIL) {
		ERROR("could not initialize incoming number %s on device %s\n", number, m_devicename);
	}
	m_listening_state = LISTENING_STATE_LISTENING;
	bool ok = false;
	while (!ok) {
		/* now spawn a new listener thread */
		if ( pthread_create (&t_listener, NULL, &_listen, this) != 0) {
			if (errno == EINTR) {
				ERROR("interrupted system call, retrying after 1 sec!\n");
				sleep(1);
				continue;
			}	
			ERROR("could not spawn new thread, reason: %s\n", strerror(errno));
			return NULL;
		}
		if (pthread_detach(t_listener) != 0) {
			ERROR("detaching of thread failed, reason: %s\n", strerror(errno));
		}	
		DBG("thread created and detached for number %s using device %s\n", number, m_devicename);
		ok = true;
	}	
	return &t_listener;
}

/** Thread start function. Parameter is IsdnDevice. 
 * This function calls listen() on the specified IsdnDevice. */
void* IsdnDeviceIncoming::_listen(void *dv) {
	((IsdnDeviceIncoming*)dv)->listen();
	return NULL; // ignored
}

void IsdnDeviceIncoming::listen() {
	char buf[100];
	char callednumberbuf[100];
	char fromSip[100];
	char toSip[100];
	int res = IsdnUtils::nb_read_answer(m_fd, buf, OPTION_DONT_STRIP, INCOMING_REINIT_TIMEOUT*10);
	DBG("after read: %s (%i bytes)\n", buf, strlen(buf));

	if (res == OK) {
		
		if (IsdnUtils::extractNumberOnCall(buf, callednumberbuf) == FAIL) {
			ERROR("received unknown sequence %s while waiting for call, exiting thread.\n", buf);
		}	else {
			DBG("incoming call detected from %s\n", buf);
			strcpy(fromSip, "sip:");
			strcat(fromSip, buf);
			strcat(fromSip, "@");
			strcat(fromSip, isdn_config->callerdomain);
			strcpy(toSip, "sip:");
			strcat(toSip, callednumberbuf);
			strcat(toSip, "@");
			strcat(toSip, isdn_config->calleedomain);
			DBG("SIP \"%s\" -> \"%s\"\n", fromSip, toSip);
			m_listening_state = LISTENING_STATE_RINGING;
			reassignIncomingDevices();
			
			strcpy(m_call_id, INCOMING_KEY_PREFIX);
			strcat(m_call_id, m_devicename);
			DBG("locking mut_state...\n");
			pthread_mutex_lock(&mut_state);
			DBG("locked. triggering outgoing call...\n");
			AmRequestUAC::dialout(m_call_id, MODULE_NAME, toSip, fromSip);
			m_listening_state = LISTENING_STATE_RINGING;
			pthread_mutex_unlock(&mut_state);
			DBG("mut_state unlocked. Now waiting a max of %i seconds for OK on SIP side\n", INCOMING_CALL_ACCEPT_TIMEOUT);
			bool ringing = true;
			int count;
			while (ringing) {
				sleep (1);
				pthread_mutex_lock(&mut_state);
				if (m_listening_state != LISTENING_STATE_RINGING) {
					// not our turn any more
					pthread_mutex_unlock(&mut_state);
					return;
				}	
				res = IsdnUtils::nb_read_answer(m_fd, buf, NO_OPTIONS, 0);
				pthread_mutex_unlock(&mut_state);
				if (res == OK) {
					DBG("read %i bytes from ttyI device: %s\n", strlen(buf), buf);
					if (strlen(buf) < 4) {
						ERROR("unknown sequence (too short)\n");
						continue;
					}	
					DBG("cmp res %i\n", strncmp("RUNG", buf, 4));
					if (strncmp("RUNG", buf, 4) == 0) {
						DBG("PSTN hangup (RUNG), canceling SIP call...\n");
						AmRequestUAC::cancel(m_cmd);
						/* a simple return is ok, device gets reinitialized from onError */
						return;
					}	
				} else {
					DBG("nothing to read %i/%i!\n", count, INCOMING_CALL_ACCEPT_TIMEOUT);
					if (count++ == INCOMING_CALL_ACCEPT_TIMEOUT)
						ringing = false;
				}
			}	
/*			pthread_mutex_lock(&mut_state);
			if (m_listening_state == LISTENING_STATE_RINGING) {
				DBG("call was NOT answered, now I'd really like to stop the SIP request ;-)\n");
				ERROR("NOT FOR PRODUCTIVE USE!!! wrong call connection IS possible!\n");
				m_listening_state = LISTENING_STATE_IDLE;
			} else {
				DBG("call was answered, someone else took care of ourselves ;-) exiting without any resets or unlocks...\n");
				DBG("just unlocking mut_state of course\n");
				pthread_mutex_unlock(&mut_state);
				return;
			}
			pthread_mutex_unlock(&mut_state);*/
		}	
	} else {
		DBG("thread for %s now exiting (timeout).\n", m_devicename);
	}
	
	initialize();
	unlock();
	m_listening_state = LISTENING_STATE_IDLE;
	reassignIncomingDevices();
}

/** Call this method id ATA failed to unlock and reuse the device. */
void IsdnDeviceIncoming::resetFailedIncomingDevice() {
	hangup();
	initialize();
	unlock();
	m_listening_state = LISTENING_STATE_IDLE;
	reassignIncomingDevices();
}

/** Threadsafe getter for listening state, uses mut_state */
int IsdnDeviceIncoming::acceptCallForKey(const char* key) {
	/* does the key match ? */
	if (strcmp(key, m_call_id) == 0) {
		DBG("found matching key %s, setting state to call active, if not yet expired\n", key);
		pthread_mutex_lock(&mut_state);
		if (m_listening_state == LISTENING_STATE_RINGING) {
			DBG("call is still pending :-)\n");
//			m_listening_state = LISTENING_STATE_CONNECTED;
			pthread_mutex_unlock(&mut_state);
			return OK;
		} else {	
			DBG("call is not pending, device is resetted already, unlocking and leaving with FAIL\n");
			pthread_mutex_unlock(&mut_state);
			return FAIL;
		}	
	} else {
		DBG("keys not matching %s <-> %s\n", key, m_devicename);
		return FAIL;
	}	
}

/** If a matching pending device is found, return it, else return NULL. */
IsdnDeviceIncoming* IsdnDeviceIncoming::getPendingCallDevice(const char* key) {
	struct list_t* cur = isdn_config->i4l_indevices;
	while (cur != NULL) {
		if (((IsdnDeviceIncoming*)cur->link)->acceptCallForKey(key) == OK) {
			DBG("success, returning matching IsdnDevice\n");
			return (IsdnDeviceIncoming*)cur->link;
		}	
		cur = cur->next;
	}	
	return NULL;
}

