/*
 * $Id: IsdnGatewayDialog.cpp,v 1.3 2003/11/16 19:28:35 ullstar Exp $
 *
 * Copyright (C) 2003 Ulrich Abend
 *
 * This file is part of isdngw, a free isdn gateway plugin 
 * for sems, a free SIP media server.
 *
 * isdngw is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * isdngw is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "globals.h"
#include "IsdnGatewayDialog.h"
#include "IsdnConfiguration.h"
#include "IsdnDevice.h"
#include "IsdnChat.h"
#include "AmAudioIsdn.h"
#include "AmApi.h"
#include "AmSession.h"
#include "AmConfig.h"
#include "log.h"

/** Creates a new IsdnGatewayDialog. The member variable m_idev is set. */
IsdnGatewayDialog::IsdnGatewayDialog(IsdnDevice *idev) 
	 : m_idev (idev), m_connectiontype(CONNECTIONTYPE_I4L) 
{}

#ifdef USE_CAPI20
/** Creates a new IsdnGatewayDialog for a CAPI connection. The controller to be used is set. */
IsdnGatewayDialog::IsdnGatewayDialog(int controller) 
	 : m_controller (controller), m_connectiontype(CONNECTIONTYPE_CAPI) 
{
	m_ccon = new IsdnCapiConnection(controller);
}
#endif

/** A connection exists. This method activates the voice 
 * connection and shuts down the call to both sides on hangup 
 */
void IsdnGatewayDialog::onSessionStart(AmRequest* req)
{
	if (m_connectiontype == CONNECTIONTYPE_I4L) {
		if (m_idev->connectVoice() == FAIL) {
			throw AmSession::Exception(500,"isdngw: sorry, error intitiating voice transmission.");
		}	

		/* SUCCESS: voice connection established, proceed */
	
		/* takes isdn as input an rtp as output */
	 	AmAudioIsdn isdn2rtp(m_idev, true);
		/* takes rtp as input an isdn as output */
		AmAudioIsdn rtp2isdn(m_idev, false);

		/* setup full duplex audio connection */
		getSession()->rtp_str.duplex(&isdn2rtp, &rtp2isdn);

		if (m_idev->getHangupRequested() == OK) {
			req->bye();
		}	
		hup(req);
	} else { /* CAPI CONNECTION */
		#ifdef USE_CAPI20
		AmAudioIsdn isdn2rtp(m_ccon, true);
		AmAudioIsdn rtp2isdn(m_ccon, false);
		
		/* setup full duplex audio connection */
		getSession()->rtp_str.duplex(&isdn2rtp, &rtp2isdn);
		#else
		ERROR("no capi support, should not get here (03)\n");
		#endif
	}	
}	

/** pre call setup */
void IsdnGatewayDialog::onBeforeCallAccept(AmRequest* req)
{
	DBG("onBeforeCallAccept called\n");
	if (m_connectiontype == CONNECTIONTYPE_I4L)
		m_idev->onBeforeCallAccept(req);
	else {
		#ifdef USE_CAPI20
		m_ccon->onBeforeCallAccept(req);
		#else
		ERROR("isdngw: no capi support, should not get here!\n");
		#endif
	}	
}

/** error in call setup */
void IsdnGatewayDialog::onError(unsigned int code, const string& reason)
{
	if (m_connectiontype == CONNECTIONTYPE_I4L) {
		DBG("onError called - calling idev's error handler\n");
		m_idev->onError(code, reason);
	} else { /* CAPI CONNECTION */
		#ifdef USE_CAPI20
		ERROR("TODO: onError for capi unimplemented\n");
		#else
		ERROR("no capi support, should not get here (04)\n");
		#endif
	}	

}

/** An active connection is being terminated from SIP side. The voice 
 * connection must be terminated, a flag is being set to show that. 
 */
void IsdnGatewayDialog::onBye(AmRequest* req)
{
	if (m_connectiontype == CONNECTIONTYPE_I4L) {
		INFO("#### now calling onBye!\n");
		m_idev->onBye();
		//hup(req);
	} else {	
		#ifdef USE_CAPI20
		DBG("onBye for capi connection, calling hangup\n");
		m_ccon->hangup();
		req->bye();
		ERROR("TODO: onBye for capi unimplemented\n");
		#else
		ERROR("no capi support, should not get here (05)\n");
		#endif
	}	
}

/** Hangs up the PSTN call and reintializes the isdn device. 
 * The lock on the device is removed. 
 */
void IsdnGatewayDialog::hup(AmRequest* req)
{

	m_idev->hangup();
	if (m_idev->initialize() == FAIL) {
		ERROR("####### initialization failed, lost device, keep it locked\n");
	} else {	
		m_idev->unlock();	
	}	
}

int IsdnGatewayDialog::onOther(AmSessionEvent* event) {
	ERROR("unimplemented onOther!!!\n");
	return 0;
}



/** Destructor. Unused by now. */
IsdnGatewayDialog::~IsdnGatewayDialog()
{
	#ifdef USE_CAPI20
	delete(m_ccon);
	#endif
}

